package org.eclipse.jdt.internal.compiler.ast;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.LocalDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Statement;
import org.eclipse.jdt.internal.compiler.codegen.BranchLabel;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.LoopingFlowContext;
import org.eclipse.jdt.internal.compiler.flow.UnconditionalFlowInfo;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;

public class ForStatement extends Statement {
  public Statement[] initializations;
  
  public Expression condition;
  
  public Statement[] increments;
  
  public Statement action;
  
  public BlockScope scope;
  
  private BranchLabel breakLabel;
  
  private BranchLabel continueLabel;
  
  int preCondInitStateIndex = -1;
  
  int preIncrementsInitStateIndex = -1;
  
  int condIfTrueInitStateIndex = -1;
  
  int mergedInitStateIndex = -1;
  
  public ForStatement(Statement[] initializations, Expression condition, Statement[] increments, Statement action, boolean neededScope, int s, int e) {
    this.sourceStart = s;
    this.sourceEnd = e;
    this.initializations = initializations;
    this.condition = condition;
    this.increments = increments;
    this.action = action;
    if (action instanceof org.eclipse.jdt.internal.compiler.ast.EmptyStatement)
      action.bits |= 0x1; 
    if (neededScope)
      this.bits |= 0x20000000; 
  }
  
  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo) {
    FlowInfo flowInfo1;
    LoopingFlowContext loopingContext;
    UnconditionalFlowInfo actionInfo;
    this.breakLabel = new BranchLabel();
    this.continueLabel = new BranchLabel();
    int initialComplaintLevel = ((flowInfo.reachMode() & 0x3) != 0) ? 1 : 0;
    if (this.initializations != null)
      for (int i = 0, count = this.initializations.length; i < count; i++)
        flowInfo = this.initializations[i].analyseCode(this.scope, flowContext, flowInfo);  
    this.preCondInitStateIndex = 
      currentScope.methodScope().recordInitializationStates(flowInfo);
    Constant cst = (this.condition == null) ? null : this.condition.constant;
    boolean isConditionTrue = !(cst != null && (cst == Constant.NotAConstant || !cst.booleanValue()));
    boolean isConditionFalse = (cst != null && cst != Constant.NotAConstant && !cst.booleanValue());
    cst = (this.condition == null) ? null : this.condition.optimizedBooleanConstant();
    boolean isConditionOptimizedTrue = !(cst != null && (cst == Constant.NotAConstant || !cst.booleanValue()));
    boolean isConditionOptimizedFalse = (cst != null && cst != Constant.NotAConstant && !cst.booleanValue());
    LoopingFlowContext condLoopContext = null;
    UnconditionalFlowInfo unconditionalFlowInfo1 = flowInfo.nullInfoLessUnconditionalCopy();
    if (this.condition != null && 
      !isConditionTrue) {
      flowInfo1 = 
        this.condition.analyseCode(
          this.scope, 
          (FlowContext)(condLoopContext = 
          new LoopingFlowContext(flowContext, flowInfo, (ASTNode)this, null, 
            null, (Scope)this.scope, true)), 
          (FlowInfo)unconditionalFlowInfo1);
      this.condition.checkNPEbyUnboxing(currentScope, flowContext, flowInfo);
    } 
    if (this.action == null || (
      this.action.isEmptyBlock() && (currentScope.compilerOptions()).complianceLevel <= 3080192L)) {
      if (condLoopContext != null)
        condLoopContext.complainOnDeferredFinalChecks(this.scope, flowInfo1); 
      if (isConditionTrue) {
        if (condLoopContext != null)
          condLoopContext.complainOnDeferredNullChecks(currentScope, 
              flowInfo1); 
        return (FlowInfo)FlowInfo.DEAD_END;
      } 
      if (isConditionFalse)
        this.continueLabel = null; 
      actionInfo = flowInfo1.initsWhenTrue().unconditionalCopy();
      loopingContext = 
        new LoopingFlowContext(flowContext, flowInfo, (ASTNode)this, 
          this.breakLabel, this.continueLabel, (Scope)this.scope, false);
    } else {
      loopingContext = 
        new LoopingFlowContext(flowContext, flowInfo, (ASTNode)this, this.breakLabel, 
          this.continueLabel, (Scope)this.scope, true);
      FlowInfo initsWhenTrue = flowInfo1.initsWhenTrue();
      this.condIfTrueInitStateIndex = 
        currentScope.methodScope().recordInitializationStates(initsWhenTrue);
      if (isConditionFalse) {
        actionInfo = FlowInfo.DEAD_END;
      } else {
        actionInfo = initsWhenTrue.unconditionalCopy();
        if (isConditionOptimizedFalse)
          actionInfo.setReachMode(1); 
      } 
      if (this.action.complainIfUnreachable((FlowInfo)actionInfo, this.scope, initialComplaintLevel, true) < 2)
        actionInfo = this.action.analyseCode(this.scope, (FlowContext)loopingContext, (FlowInfo)actionInfo).unconditionalInits(); 
      if ((actionInfo.tagBits & 
        loopingContext.initsOnContinue.tagBits & 
        0x1) != 0) {
        this.continueLabel = null;
      } else {
        if (condLoopContext != null)
          condLoopContext.complainOnDeferredFinalChecks(this.scope, 
              flowInfo1); 
        actionInfo = actionInfo.mergedWith(loopingContext.initsOnContinue);
        loopingContext.complainOnDeferredFinalChecks(this.scope, 
            (FlowInfo)actionInfo);
      } 
    } 
    FlowInfo exitBranch = flowInfo.copy();
    LoopingFlowContext incrementContext = null;
    if (this.continueLabel != null) {
      if (this.increments != null) {
        FlowInfo flowInfo2;
        incrementContext = 
          new LoopingFlowContext(flowContext, flowInfo, (ASTNode)this, null, 
            null, (Scope)this.scope, true);
        UnconditionalFlowInfo unconditionalFlowInfo = actionInfo;
        this.preIncrementsInitStateIndex = 
          currentScope.methodScope().recordInitializationStates((FlowInfo)unconditionalFlowInfo);
        for (int i = 0, count = this.increments.length; i < count; i++)
          flowInfo2 = this.increments[i]
            .analyseCode(this.scope, (FlowContext)incrementContext, (FlowInfo)unconditionalFlowInfo); 
        incrementContext.complainOnDeferredFinalChecks(this.scope, 
            (FlowInfo)(actionInfo = flowInfo2.unconditionalInits()));
      } 
      exitBranch.addPotentialInitializationsFrom((FlowInfo)actionInfo)
        .addInitializationsFrom(flowInfo1.initsWhenFalse());
    } else {
      exitBranch.addInitializationsFrom(flowInfo1.initsWhenFalse());
      if (this.increments != null && 
        initialComplaintLevel == 0)
        currentScope.problemReporter().fakeReachable((ASTNode)this.increments[0]); 
    } 
    if (condLoopContext != null)
      condLoopContext.complainOnDeferredNullChecks(currentScope, 
          (FlowInfo)actionInfo); 
    loopingContext.complainOnDeferredNullChecks(currentScope, 
        (FlowInfo)actionInfo);
    if (incrementContext != null)
      incrementContext.complainOnDeferredNullChecks(currentScope, 
          (FlowInfo)actionInfo); 
    if (loopingContext.hasEscapingExceptions()) {
      UnconditionalFlowInfo unconditionalFlowInfo;
      FlowInfo loopbackFlowInfo = flowInfo.copy();
      if (this.continueLabel != null)
        unconditionalFlowInfo = loopbackFlowInfo.mergedWith(loopbackFlowInfo.unconditionalCopy().addNullInfoFrom((FlowInfo)actionInfo).unconditionalInits()); 
      loopingContext.simulateThrowAfterLoopBack((FlowInfo)unconditionalFlowInfo);
    } 
    UnconditionalFlowInfo unconditionalFlowInfo2 = FlowInfo.mergedOptimizedBranches(
        ((loopingContext.initsOnBreak.tagBits & 
        0x3) != 0) ? 
        (FlowInfo)loopingContext.initsOnBreak : 
        flowInfo.addInitializationsFrom((FlowInfo)loopingContext.initsOnBreak), 
        isConditionOptimizedTrue, 
        exitBranch, 
        isConditionOptimizedFalse, 
        !isConditionTrue);
    if (this.initializations != null)
      for (int i = 0; i < this.initializations.length; i++) {
        Statement init = this.initializations[i];
        if (init instanceof LocalDeclaration) {
          LocalVariableBinding binding = ((LocalDeclaration)init).binding;
          unconditionalFlowInfo2.resetAssignmentInfo(binding);
        } 
      }  
    this.mergedInitStateIndex = currentScope.methodScope().recordInitializationStates((FlowInfo)unconditionalFlowInfo2);
    return (FlowInfo)unconditionalFlowInfo2;
  }
  
  public void generateCode(BlockScope currentScope, CodeStream codeStream) {
    if ((this.bits & Integer.MIN_VALUE) == 0)
      return; 
    int pc = codeStream.position;
    if (this.initializations != null)
      for (int i = 0, max = this.initializations.length; i < max; i++)
        this.initializations[i].generateCode(this.scope, codeStream);  
    Constant cst = (this.condition == null) ? null : this.condition.optimizedBooleanConstant();
    boolean isConditionOptimizedFalse = (cst != null && cst != Constant.NotAConstant && !cst.booleanValue());
    if (isConditionOptimizedFalse) {
      this.condition.generateCode(this.scope, codeStream, false);
      if ((this.bits & 0x20000000) != 0)
        codeStream.exitUserScope(this.scope); 
      if (this.mergedInitStateIndex != -1) {
        codeStream.removeNotDefinitelyAssignedVariables((Scope)currentScope, this.mergedInitStateIndex);
        codeStream.addDefinitelyAssignedVariables((Scope)currentScope, this.mergedInitStateIndex);
      } 
      codeStream.recordPositionsFrom(pc, this.sourceStart);
      return;
    } 
    BranchLabel actionLabel = new BranchLabel(codeStream);
    actionLabel.tagBits |= 0x2;
    BranchLabel conditionLabel = new BranchLabel(codeStream);
    this.breakLabel.initialize(codeStream);
    if (this.continueLabel == null) {
      conditionLabel.place();
      if (this.condition != null && this.condition.constant == Constant.NotAConstant)
        this.condition.generateOptimizedBoolean(this.scope, codeStream, null, this.breakLabel, true); 
    } else {
      this.continueLabel.initialize(codeStream);
      if (this.condition != null && 
        this.condition.constant == Constant.NotAConstant && ((
        this.action != null && !this.action.isEmptyBlock()) || this.increments != null)) {
        conditionLabel.tagBits |= 0x2;
        int jumpPC = codeStream.position;
        codeStream.goto_(conditionLabel);
        codeStream.recordPositionsFrom(jumpPC, this.condition.sourceStart);
      } 
    } 
    if (this.action != null) {
      if (this.condIfTrueInitStateIndex != -1)
        codeStream.addDefinitelyAssignedVariables(
            (Scope)currentScope, 
            this.condIfTrueInitStateIndex); 
      actionLabel.place();
      this.action.generateCode(this.scope, codeStream);
    } else {
      actionLabel.place();
    } 
    if (this.preIncrementsInitStateIndex != -1) {
      codeStream.removeNotDefinitelyAssignedVariables((Scope)currentScope, this.preIncrementsInitStateIndex);
      codeStream.addDefinitelyAssignedVariables((Scope)currentScope, this.preIncrementsInitStateIndex);
    } 
    if (this.continueLabel != null) {
      this.continueLabel.place();
      if (this.increments != null)
        for (int i = 0, max = this.increments.length; i < max; i++)
          this.increments[i].generateCode(this.scope, codeStream);  
      if (this.preCondInitStateIndex != -1)
        codeStream.removeNotDefinitelyAssignedVariables((Scope)currentScope, this.preCondInitStateIndex); 
      conditionLabel.place();
      if (this.condition != null && this.condition.constant == Constant.NotAConstant) {
        this.condition.generateOptimizedBoolean(this.scope, codeStream, actionLabel, null, true);
      } else {
        codeStream.goto_(actionLabel);
      } 
    } else if (this.preCondInitStateIndex != -1) {
      codeStream.removeNotDefinitelyAssignedVariables((Scope)currentScope, this.preCondInitStateIndex);
    } 
    if ((this.bits & 0x20000000) != 0)
      codeStream.exitUserScope(this.scope); 
    if (this.mergedInitStateIndex != -1) {
      codeStream.removeNotDefinitelyAssignedVariables((Scope)currentScope, this.mergedInitStateIndex);
      codeStream.addDefinitelyAssignedVariables((Scope)currentScope, this.mergedInitStateIndex);
    } 
    this.breakLabel.place();
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }
  
  public StringBuffer printStatement(int tab, StringBuffer output) {
    printIndent(tab, output).append("for (");
    if (this.initializations != null)
      for (int i = 0; i < this.initializations.length; i++) {
        if (i > 0)
          output.append(", "); 
        this.initializations[i].print(0, output);
      }  
    output.append("; ");
    if (this.condition != null)
      this.condition.printExpression(0, output); 
    output.append("; ");
    if (this.increments != null)
      for (int i = 0; i < this.increments.length; i++) {
        if (i > 0)
          output.append(", "); 
        this.increments[i].print(0, output);
      }  
    output.append(") ");
    if (this.action == null) {
      output.append(';');
    } else {
      output.append('\n');
      this.action.printStatement(tab + 1, output);
    } 
    return output;
  }
  
  public void resolve(BlockScope upperScope) {
    this.scope = ((this.bits & 0x20000000) != 0) ? new BlockScope(upperScope) : upperScope;
    if (this.initializations != null)
      for (int i = 0, length = this.initializations.length; i < length; i++)
        this.initializations[i].resolve(this.scope);  
    if (this.condition != null) {
      TypeBinding type = this.condition.resolveTypeExpecting(this.scope, (TypeBinding)TypeBinding.BOOLEAN);
      this.condition.computeConversion((Scope)this.scope, type, type);
    } 
    if (this.increments != null)
      for (int i = 0, length = this.increments.length; i < length; i++)
        this.increments[i].resolve(this.scope);  
    if (this.action != null)
      this.action.resolve(this.scope); 
  }
  
  public void traverse(ASTVisitor visitor, BlockScope blockScope) {
    if (visitor.visit(this, blockScope)) {
      if (this.initializations != null) {
        int initializationsLength = this.initializations.length;
        for (int i = 0; i < initializationsLength; i++)
          this.initializations[i].traverse(visitor, this.scope); 
      } 
      if (this.condition != null)
        this.condition.traverse(visitor, this.scope); 
      if (this.increments != null) {
        int incrementsLength = this.increments.length;
        for (int i = 0; i < incrementsLength; i++)
          this.increments[i].traverse(visitor, this.scope); 
      } 
      if (this.action != null)
        this.action.traverse(visitor, this.scope); 
    } 
    visitor.endVisit(this, blockScope);
  }
  
  public boolean doesNotCompleteNormally() {
    Constant cst = (this.condition == null) ? null : this.condition.constant;
    boolean isConditionTrue = !(cst != null && (cst == Constant.NotAConstant || !cst.booleanValue()));
    cst = (this.condition == null) ? null : this.condition.optimizedBooleanConstant();
    boolean isConditionOptimizedTrue = (cst == null) ? true : ((cst != Constant.NotAConstant && cst.booleanValue()));
    return ((isConditionTrue || isConditionOptimizedTrue) && (this.action == null || !this.action.breaksOut(null)));
  }
  
  public boolean completesByContinue() {
    return this.action.continuesAtOuterLabel();
  }
}
